"""Collection of SDMX-JSON schemas for reference metadata queries."""

from typing import Sequence

from msgspec import Struct

from pysdmx.fmr.reader import _merge_attributes
from pysdmx.model import MetadataReport


class JsonMetadataSets(Struct, frozen=True):
    """SDMX-JSON payload for the list of metadata sets."""

    metadataSets: Sequence[MetadataReport]


class JsonMetadataMessage(Struct, frozen=True):
    """SDMX-JSON payload for /metadata queries."""

    data: JsonMetadataSets

    def to_model(self) -> MetadataReport:
        """Returns the requested metadata report."""
        r = self.data.metadataSets[0]
        attrs = _merge_attributes(r.attributes)
        return MetadataReport(r.id, r.name, r.metadataflow, r.targets, attrs)



"""Collection of SDMX-JSON schemas for codes and codelists."""

from datetime import datetime, timezone as tz
from typing import Optional, Sequence, Tuple

from msgspec import Struct

from pysdmx.fmr.sdmx.core import JsonAnnotation
from pysdmx.model import Code, Codelist, HierarchicalCode, Hierarchy
from pysdmx.util import parse_item_urn


class JsonCode(Struct, frozen=True):
    """SDMX-JSON payload for codes."""

    id: str
    annotations: Sequence[JsonAnnotation] = ()
    name: Optional[str] = None
    description: Optional[str] = None

    def __handle_date(self, datestr: str) -> datetime:
        return datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%S%z")

    def __get_val(
        self, a: JsonAnnotation
    ) -> Tuple[Optional[datetime], Optional[datetime]]:
        vals = a.title.split("/")
        if a.title.startswith("/"):
            return (None, self.__handle_date(vals[1]))
        else:
            valid_from = self.__handle_date(vals[0])
            valid_to = self.__handle_date(vals[1]) if vals[1] else None
            return (valid_from, valid_to)

    def to_model(self) -> Code:
        """Converts a JsonCode to a standard code."""
        vp = [a for a in self.annotations if a.type == "FR_VALIDITY_PERIOD"]
        vf, vt = self.__get_val(vp[0]) if vp else (None, None)
        return Code(self.id, self.name, self.description, vf, vt)


class JsonCodelist(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a codelist."""

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    codes: Sequence[JsonCode] = ()

    def to_model(self) -> Codelist:
        """Converts a JsonCodelist to a standard codelist."""
        return Codelist(
            self.id,
            self.name,
            self.agency,
            self.description,
            self.version,
            [i.to_model() for i in self.codes],
        )


class JsonValuelist(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a valuelist."""

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    valueItems: Sequence[JsonCode] = ()

    def to_model(self) -> Codelist:
        """Converts a JsonValuelist to a standard codelist."""
        return Codelist(
            self.id,
            self.name,
            self.agency,
            self.description,
            self.version,
            [i.to_model() for i in self.valueItems],
        )


class JsonCodelists(Struct, frozen=True):
    """SDMX-JSON payload for lists of codes."""

    codelists: Sequence[JsonCodelist] = ()
    valuelists: Sequence[JsonValuelist] = ()


class JsonCodelistMessage(Struct, frozen=True):
    """SDMX-JSON payload for /codelist queries."""

    data: JsonCodelists

    def to_model(self) -> Codelist:
        """Returns the requested codelist."""
        if self.data.codelists:
            return self.data.codelists[0].to_model()
        else:
            return self.data.valuelists[0].to_model()


class JsonHierarchicalCode(Struct, frozen=True):
    """Fusion-JSON payload for hierarchical codes."""

    code: str
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    hierarchicalCodes: Sequence["JsonHierarchicalCode"] = ()

    def __find_code(self, codelists: Sequence[Codelist], urn: str) -> Code:
        r = parse_item_urn(urn)
        f = [
            c
            for c in codelists
            if (
                c.agency == r.agency
                and c.id == r.id
                and c.version == r.version
            )
        ]
        return [c for c in f[0].codes if c.id == r.item_id][0]

    def to_model(self, codelists: Sequence[Codelist]) -> HierarchicalCode:
        """Converts a JsonHierarchicalCode to a hierachical code."""
        code = self.__find_code(codelists, self.code)
        codes = [c.to_model(codelists) for c in self.hierarchicalCodes]
        vf = self.validFrom.replace(tzinfo=tz.utc) if self.validFrom else None
        vt = self.validTo.replace(tzinfo=tz.utc) if self.validTo else None
        return HierarchicalCode(
            code.id,
            code.name,
            code.description,
            code.valid_from,
            code.valid_to,
            vf,
            vt,
            codes,
        )


class JsonHierarchy(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a hierarchy."""

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    hierarchicalCodes: Sequence[JsonHierarchicalCode] = ()

    def to_model(self, codelists: Sequence[Codelist]) -> Hierarchy:
        """Converts a JsonHierarchy to a standard hierarchy."""
        return Hierarchy(
            self.id,
            self.name,
            self.agency,
            self.description,
            self.version,
            [i.to_model(codelists) for i in self.hierarchicalCodes],
        )


class JsonHierarchies(Struct, frozen=True):
    """SDMX-JSON payload for hierarchies."""

    codelists: Sequence[JsonCodelist] = ()
    hierarchies: Sequence[JsonHierarchy] = ()

    def to_model(self) -> Hierarchy:
        """Returns the requested hierarchy."""
        cls = [cl.to_model() for cl in self.codelists]
        return self.hierarchies[0].to_model(cls)


class JsonHierarchyMessage(Struct, frozen=True):
    """SDMX-JSON payload for /hierarchy queries."""

    data: JsonHierarchies

    def to_model(self) -> Hierarchy:
        """Returns the requested hierarchy."""
        return self.data.to_model()



"""SDMX simplified **domain model**, expressed as **Python classes**.

This module contains data classes representing a **simplified and opinionated
subset** of the SDMX information model.
"""

from re import Pattern
from typing import Any

from pysdmx.model.category import Category, CategoryScheme
from pysdmx.model.code import Code, Codelist, HierarchicalCode, Hierarchy
from pysdmx.model.concept import Concept, ConceptScheme, DataType, Facets
from pysdmx.model.dataflow import (
    ArrayBoundaries,
    Component,
    Components,
    DataflowInfo,
    Role,
    Schema,
)
from pysdmx.model.map import (
    ComponentMap,
    DatePatternMap,
    FixedValueMap,
    ImplicitComponentMap,
    MultiComponentMap,
    MultiValueMap,
    StructureMap,
    ValueMap,
)
from pysdmx.model.metadata import MetadataAttribute, MetadataReport
from pysdmx.model.organisation import Contact, DataflowRef, Organisation


def encoders(obj: Any) -> Any:
    """Encoders for msgspec serialization.

    The pysdmx model classes are based on msgspec `Struct` classes. These
    classes offer serialization and deserialization into various formats
    for many Python types. However, some types used in pysdmx are not
    supported. This will lead to serialization errors, unless `encoders`
    is used. See details at https://jcristharif.com/msgspec/extending.html.

    For example:

        import msgspec
        from pysdmx.model import encoders
        encoder = msgspec.json.Encoder(enc_hook=encoders)

    Args:
        obj: The object to be encoded

    Returns:
        The received object converted to supported Python types

    Raises:
        NotImplementedError: In case the object type is not one of the types
            needing conversion
    """
    if isinstance(obj, Pattern):
        return f"regex:{obj.pattern}"
    elif isinstance(obj, Components):
        return list(obj)
    else:
        # Raise a NotImplementedError for other types
        raise NotImplementedError(
            f"Objects of type {type(obj)} are not supported"
        )


__all__ = [
    "ArrayBoundaries",
    "Category",
    "CategoryScheme",
    "Code",
    "Codelist",
    "Component",
    "Components",
    "ComponentMap",
    "Concept",
    "ConceptScheme",
    "Contact",
    "DataflowInfo",
    "DataflowRef",
    "DataType",
    "DatePatternMap",
    "Facets",
    "HierarchicalCode",
    "Hierarchy",
    "ImplicitComponentMap",
    "StructureMap",
    "MetadataAttribute",
    "MetadataReport",
    "MultiComponentMap",
    "MultiValueMap",
    "Organisation",
    "Role",
    "Schema",
    "ValueMap",
    "FixedValueMap",
]



"""Types of errors that can be returned by pysdmx."""

from typing import Optional


class Error(Exception):
    """Base ``pysdmx`` error class.

    This class is inherited by more specific error classes and so, ideally,
    you should not encounter it directly when calling library functions.
    """

    def __init__(
        self, status: int, title: str, description: Optional[str] = None
    ):
        """Instantiates a new error, with the supplied paramaters."""
        self.status = status
        self.title = title
        self.description = description
        super().__init__(self.description)


class RetriableError(Error):
    """A type of errors that may be resolved after retrying.

    This class is inherited by more specific error classes and so, ideally,
    you should not encounter it directly when calling library functions.
    """


class Unavailable(RetriableError):
    """The targeted service is not available.

    This can be for a variety of reasons such as:

    - A network issue between the client making the call and the targeted
      service
    - The targeted service is temporarily unavailable
    - The targeted service is overloaded and therefore unable to process
      the request

    This type of errors is considered as **retriable** and so clients can
    retry executing the query at a later stage.
    """


class ClientError(Error):
    """The request from the client is **invalid**.

    This type of errors is considered as **non-retriable** and so clients
    should **not** retry the query before investigating the issue first.
    """


class ServiceError(Error):
    """The targeted service reported an error.

    This type of errors is considered as **non-retriable** and so clients
    should **not** retry the query before investigating the issue first.
    """


class NotFound(Error):
    """The requested resources are **not available**.

    This can be for a variety of reasons such as:

    - There is a typo in the parameters passed to the function
    - The resources exist but in another service than the targeted one

    This type of errors is considered as **non-retriable** and so clients
    should **not** retry the query before investigating the issue first.
    """



"""Collection of Fusion-JSON schemas for SDMX-REST schema queries."""
from typing import List, Sequence

from msgspec import Struct

from pysdmx.fmr.fusion.code import FusionCodelist
from pysdmx.fmr.fusion.concept import FusionConceptScheme
from pysdmx.fmr.fusion.constraint import FusionContentConstraint
from pysdmx.fmr.fusion.dsd import FusionDataStructure
from pysdmx.model import Schema


class FusionLink(Struct, frozen=True):
    """Fusion-JSON payload for link objects."""

    urn: str


class FusionHeader(Struct, frozen=True):
    """Fusion-JSON payload for message header."""

    links: Sequence[FusionLink] = ()


class FusionSchemaMessage(
    Struct,
    frozen=True,
):
    """Fusion-JSON payload for /schema queries."""

    meta: FusionHeader
    ConceptScheme: Sequence[FusionConceptScheme]
    DataStructure: Sequence[FusionDataStructure]
    ValueList: Sequence[FusionCodelist] = ()
    Codelist: Sequence[FusionCodelist] = ()
    DataConstraint: Sequence[FusionContentConstraint] = ()

    def to_model(
        self,
        context: str,
        agency: str,
        id_: str,
        version: str,
    ) -> Schema:
        """Returns the requested schema."""
        cls: List[FusionCodelist] = []
        cls.extend(self.Codelist)
        cls.extend(self.ValueList)
        components = self.DataStructure[0].get_components(
            self.ConceptScheme, cls, self.DataConstraint
        )
        urns = [a.urn for a in self.meta.links]
        return Schema(context, agency, id_, components, version, urns)



"""Collection of Fusion-JSON schemas for concepts."""

from typing import Optional, Sequence

from msgspec import Struct

from pysdmx.fmr.fusion.code import FusionCodelist
from pysdmx.fmr.fusion.core import FusionRepresentation, FusionString
from pysdmx.model.concept import Concept, ConceptScheme as CS, DataType


class FusionConcept(Struct, frozen=True):
    """Fusion-JSON payload for concepts."""

    id: str
    names: Sequence[FusionString]
    representation: Optional[FusionRepresentation] = None
    descriptions: Optional[Sequence[FusionString]] = None

    def to_model(self, codelists: Sequence[FusionCodelist]) -> Concept:
        """Converts a FusionConcept to a standard concept."""
        dt = (
            DataType(self.representation.textFormat.textType)
            if self.representation and self.representation.textFormat
            else DataType.STRING
        )
        f = self.representation.to_facets() if self.representation else None
        c = (
            self.representation.to_enumeration(codelists, [])
            if self.representation
            else []
        )
        d = self.descriptions[0].value if self.descriptions else None
        cl_ref = (
            self.representation.representation
            if self.representation and c
            else None
        )
        return Concept(
            self.id,
            dtype=dt,
            facets=f,
            name=self.names[0].value,
            description=d,
            codes=c,
            enum_ref=cl_ref,
        )


class FusionConceptScheme(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a concept scheme."""

    id: str
    names: Sequence[FusionString]
    agency: str
    descriptions: Optional[Sequence[FusionString]] = None
    version: str = "1.0"
    items: Sequence[FusionConcept] = ()

    def to_model(self, codelists: Sequence[FusionCodelist]) -> CS:
        """Converts a FusionConceptScheme to a standard concept scheme."""
        d = self.descriptions[0].value if self.descriptions else None
        return CS(
            self.id,
            self.names[0].value,
            self.agency,
            d,
            self.version,
            [c.to_model(codelists) for c in self.items],
        )


class FusionConcepSchemeMessage(
    Struct,
    frozen=True,
):
    """Fusion-JSON payload for /conceptscheme queries."""

    Codelist: Sequence[FusionCodelist]
    ConceptScheme: Sequence[FusionConceptScheme]

    def to_model(self) -> CS:
        """Returns the requested concept scheme."""
        return self.ConceptScheme[0].to_model(self.Codelist)



"""Collection of SDMX-JSON schemas for SDMX-REST schema queries."""
from typing import Sequence

from msgspec import Struct

from pysdmx.fmr.sdmx.code import JsonCodelist, JsonValuelist
from pysdmx.fmr.sdmx.concept import JsonConceptScheme
from pysdmx.fmr.sdmx.constraint import JsonContentConstraint
from pysdmx.fmr.sdmx.core import JsonHeader
from pysdmx.fmr.sdmx.dsd import JsonDataStructure
from pysdmx.model import Components, Schema


class JsonSchemas(
    Struct,
    frozen=True,
):
    """SDMX-JSON payload schema structures."""

    conceptSchemes: Sequence[JsonConceptScheme]
    dataStructures: Sequence[JsonDataStructure]
    valuelists: Sequence[JsonValuelist] = ()
    codelists: Sequence[JsonCodelist] = ()
    contentConstraints: Sequence[JsonContentConstraint] = ()

    def to_model(self) -> Components:
        """Returns the requested schema."""
        cls = [cl.to_model() for cl in self.codelists]
        cls.extend([vl.to_model() for vl in self.valuelists])
        return self.dataStructures[0].dataStructureComponents.to_model(
            self.conceptSchemes, cls, self.contentConstraints
        )


class JsonSchemaMessage(
    Struct,
    frozen=True,
):
    """SDMX-JSON payload for /schema queries."""

    meta: JsonHeader
    data: JsonSchemas

    def to_model(
        self,
        context: str,
        agency: str,
        id_: str,
        version: str,
    ) -> Schema:
        """Returns the requested schema."""
        components = self.data.to_model()
        urns = [a.urn for a in self.meta.links]
        return Schema(context, agency, id_, components, version, urns)



"""Collection of SDMX-JSON schemas for concepts."""

from typing import Optional, Sequence

from msgspec import Struct

from pysdmx.fmr.sdmx.code import JsonCodelist
from pysdmx.fmr.sdmx.core import JsonRepresentation
from pysdmx.model import Concept, ConceptScheme, DataType


class JsonConcept(Struct, frozen=True):
    """SDMX-JSON payload for concepts."""

    id: str
    coreRepresentation: Optional[JsonRepresentation] = None
    name: Optional[str] = None
    description: Optional[str] = None

    def to_model(self, codelists: Sequence[JsonCodelist]) -> Concept:
        """Converts a JsonConcept to a standard concept."""
        repr_ = self.coreRepresentation
        if repr_:
            if repr_.enumerationFormat:
                dt = DataType(
                    repr_.enumerationFormat.textType,
                )
            else:
                dt = DataType(
                    repr_.format.textType,  # type: ignore[union-attr]
                )
            facets = repr_.to_facets()
            codes = repr_.to_enumeration(
                [cl.to_model() for cl in codelists], []
            )
            cl_ref = repr_.enumeration
        else:
            dt = DataType.STRING
            facets = None
            codes = []
            cl_ref = None
        return Concept(
            self.id,
            dtype=dt,
            facets=facets,
            name=self.name,
            description=self.description,
            codes=codes,
            enum_ref=cl_ref,
        )


class JsonConceptScheme(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a concept scheme."""

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    concepts: Sequence[JsonConcept] = ()

    def to_model(self, codelists: Sequence[JsonCodelist]) -> ConceptScheme:
        """Converts a JsonConceptScheme to a standard concept scheme."""
        return ConceptScheme(
            self.id,
            self.name,
            self.agency,
            self.description,
            self.version,
            [c.to_model(codelists) for c in self.concepts],
        )


class JsonConceptSchemes(
    Struct,
    frozen=True,
):
    """SDMX-JSON payload for the list of concept schemes."""

    codelists: Sequence[JsonCodelist]
    conceptSchemes: Sequence[JsonConceptScheme]


class JsonConcepSchemeMessage(
    Struct,
    frozen=True,
):
    """SDMX-JSON payload for /conceptscheme queries."""

    data: JsonConceptSchemes

    def to_model(self) -> ConceptScheme:
        """Returns the requested concept scheme."""
        return self.data.conceptSchemes[0].to_model(self.data.codelists)



"""Model for SDMX Organisations (data providers, agencies, etc).

Organisations in SDMX can play different roles, such as:

- Maintenance agencies, i.e. organisations that maintain metadata.
- Data (or metadata) providers, i.e. organisations that provide data or
  metadata.
- Data (or metadata) consumers, i.e. organisations that collect data or
  metadata.

Contact details may be included with the information about an organisation.

Currently, only agencies and providers are supported in pysdmx (as consumers
seem to be underused in SDMX currently).

Furthermore, in pysdmx, an organisation may reference a list of dataflows
it maintains (if the organisation is an agency) or for which it provides
data (if the organisation is a data provider).
"""

from typing import Optional, Sequence

from msgspec import Struct


class Contact(Struct, frozen=True, omit_defaults=True):
    """Contact details such as the name of a contact and his email address.

    Attributes:
        id: An identifier for a contact. If the contact represents a person,
            this could be the person's username in the organisation.
        name: The contact name, which could be the name of a person, the name
            of a service ("e.g. Support"), etc.
        department: The department in which the contact is located (e.g.
            "Statistics").
        role: The contact's role, which could be his job title, or a role such
            as data owner, data steward, subject matter expert, etc.
        telephones: A list of telephone numbers.
        faxes: A list of fax numbers.
        uris: A list of URLs relevant for the contact (e.g. a link to an online
            form that can be used to send questions, a link to a support forum,
            etc.).
        emails: a list of email addresses.
    """

    id: Optional[str] = None
    name: Optional[str] = None
    department: Optional[str] = None
    role: Optional[str] = None
    telephones: Optional[Sequence[str]] = None
    faxes: Optional[Sequence[str]] = None
    uris: Optional[Sequence[str]] = None
    emails: Optional[Sequence[str]] = None


class DataflowRef(Struct, frozen=True, omit_defaults=True):
    """Provide core information about a dataflow.

    Attributes:
        id: The dataflow identifier (e.g. BIS_MACRO).
        agency: The organisation (or unit) responsible for the dataflow.
        name: The dataflow name (e.g. MACRO dataflow).
        description: Additional descriptive information about the dataflow.
        version: The version of the dataflow (e.g. 1.0).
    """

    id: str
    agency: str
    name: Optional[str] = None
    description: Optional[str] = None
    version: str = "1.0"

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = self.id
        if self.name:
            out = f"{out} ({self.name})"
        return out


class Organisation(Struct, frozen=True, omit_defaults=True):
    """An organisation such as a provider of data or a metadata maintainer.

    Central Banks, International Organisations, statistical offices, market
    data vendors are typical examples of organisations participating in
    statistical data exchanges.

    Organisations may have one or more contact details.

    Attributes:
        id: The identifier of the organisation (e.g. BIS).
        name: The name of the organisation (e.g. Bank for
            International Settlements).
        description: Additional descriptive information about
            the organisation.
        contacts: Contact details (email address, support forum, etc).
        dataflows: The dataflows maintained by the organisation if it is
            a maintenance agency or for which the organisation provides data
            if it is a data provider.
    """

    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    contacts: Sequence[Contact] = ()
    dataflows: Sequence[DataflowRef] = ()

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = self.id
        if self.name:
            out = f"{out} ({self.name})"
        return out

    def __hash__(self) -> int:
        """Returns the organisation's hash."""
        return hash(self.id)



"""Collection of SDMX-JSON schemas for organisations."""

from collections import defaultdict
from typing import Dict, Sequence, Set

from msgspec import Struct

from pysdmx.fmr.sdmx.pa import JsonProvisionAgreement
from pysdmx.model import DataflowRef, Organisation
from pysdmx.util import parse_urn


class JsonDataProviderScheme(Struct, frozen=True):
    """SDMX-JSON payload for a data provider scheme."""

    dataProviders: Sequence[Organisation]

    def __get_df_ref(self, ref: str) -> DataflowRef:
        a = parse_urn(ref)
        return DataflowRef(a.id, a.agency, version=a.version)

    def to_model(
        self, pas: Sequence[JsonProvisionAgreement]
    ) -> Sequence[Organisation]:
        """Converts a JsonDataProviderScheme to a list of Organisations."""
        if pas:
            paprs: Dict[str, Set[DataflowRef]] = defaultdict(set)
            for pa in pas:
                df = self.__get_df_ref(pa.structureUsage)
                pr = pa.dataProvider[pa.dataProvider.rindex(".") + 1 :]
                paprs[pr].add(df)
            return [
                Organisation(
                    p.id, p.name, p.description, p.contacts, list(paprs[p.id])
                )
                for p in self.dataProviders
            ]
        else:
            return self.dataProviders


class JsonDataProviderSchemes(Struct, frozen=True):
    """SDMX-JSON payload for the list of data provider schemes."""

    dataProviderSchemes: Sequence[JsonDataProviderScheme]
    provisionAgreements: Sequence[JsonProvisionAgreement] = ()

    def to_model(self) -> Sequence[Organisation]:
        """Converts a JsonDataProviderSchemes to a list of Organisations."""
        return self.dataProviderSchemes[0].to_model(self.provisionAgreements)


class JsonProviderMessage(Struct, frozen=True):
    """SDMX-JSON payload for /dataproviderscheme queries."""

    data: JsonDataProviderSchemes

    def to_model(self) -> Sequence[Organisation]:
        """Returns the requested list of providers."""
        return self.data.to_model()


class JsonAgencyScheme(Struct, frozen=True):
    """SDMX-JSON payload for an agency scheme."""

    agencyID: str
    agencies: Sequence[Organisation]


class JsonAgencySchemes(Struct, frozen=True):
    """SDMX-JSON payload for the list of agency schemes."""

    agencySchemes: Sequence[JsonAgencyScheme]


class JsonAgencyMessage(Struct, frozen=True):
    """SDMX-JSON payload for /agencyscheme queries."""

    data: JsonAgencySchemes

    def __add_owner(self, owner: str, a: Organisation) -> Organisation:
        oid = f"{owner}.{a.id}" if owner != "SDMX" else a.id
        return Organisation(oid, a.name, a.description, a.contacts)

    def to_model(self) -> Sequence[Organisation]:
        """Returns the requested list of agencies."""
        return [
            self.__add_owner(self.data.agencySchemes[0].agencyID, a)
            for a in self.data.agencySchemes[0].agencies
        ]



"""Collection of Fusion-JSON schemas for categories and category schemes."""

from collections import defaultdict
from typing import Dict, Optional, Sequence

from msgspec import Struct

from pysdmx.fmr.fusion.core import FusionString
from pysdmx.fmr.fusion.dataflow import FusionDataflowRef
from pysdmx.model import Category, CategoryScheme as CS, DataflowRef
from pysdmx.util import find_by_urn


class FusionCategorisation(Struct, frozen=True):
    """Fusion-JSON payload for a categorisation."""

    categoryReference: str
    structureReference: str


class FusionCategory(Struct, frozen=True):
    """Fusion-JSON payload for a category."""

    id: str
    names: Sequence[FusionString]
    descriptions: Optional[Sequence[FusionString]] = None
    items: Sequence["FusionCategory"] = ()

    def to_model(self) -> Category:
        """Converts a FusionCode to a standard code."""
        return Category(
            self.id,
            self.names[0].value,
            self.descriptions[0].value if self.descriptions else None,
            [c.to_model() for c in self.items],
        )


class FusionCategoryScheme(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a category scheme."""

    id: str
    names: Sequence[FusionString]
    agency: str
    descriptions: Optional[Sequence[FusionString]] = None
    version: str = "1.0"
    items: Sequence[FusionCategory] = ()

    def to_model(self) -> CS:
        """Converts a JsonCodelist to a standard codelist."""
        return CS(
            self.id,
            self.names[0].value,
            self.agency,
            self.descriptions[0].value if self.descriptions else None,
            self.version,
            [c.to_model() for c in self.items],
        )


class FusionCategorySchemeMessage(Struct, frozen=True):
    """Fusion-JSON payload for /categoryscheme queries."""

    Categorisation: Sequence[FusionCategorisation]
    CategoryScheme: Sequence[FusionCategoryScheme]
    Dataflow: Sequence[FusionDataflowRef]

    def __group_flows(self) -> defaultdict[str, list[DataflowRef]]:
        out: defaultdict[str, list[DataflowRef]] = defaultdict(list)
        for c in self.Categorisation:
            d = find_by_urn(self.Dataflow, c.structureReference)
            src = c.categoryReference[c.categoryReference.find(")") + 2 :]
            out[src].append(d.to_model())
        return out

    def __add_flows(
        self, cat: Category, cni: str, cf: Dict[str, list[DataflowRef]]
    ) -> None:
        if cat.categories:
            for c in cat.categories:
                self.__add_flows(c, f"{cni}.{c.id}", cf)
        if cni in cf:
            cat.dataflows = cf[cni]

    def to_model(self) -> CS:
        """Returns the requested codelist."""
        cf = self.__group_flows()
        cs = self.CategoryScheme[0].to_model()
        for c in cs:
            self.__add_flows(c, c.id, cf)
        return cs



"""Collection of SDMX-JSON schemas for SDMX-REST schema queries."""

from typing import Dict, List, Optional, Sequence, Tuple

from msgspec import Struct

from pysdmx.fmr.sdmx.concept import JsonConcept, JsonConceptScheme
from pysdmx.fmr.sdmx.constraint import JsonContentConstraint
from pysdmx.fmr.sdmx.core import JsonRepresentation
from pysdmx.model import (
    ArrayBoundaries,
    Code,
    Codelist,
    Component,
    Components,
    DataType,
    Facets,
    Role,
)
from pysdmx.util import parse_item_urn


def _find_concept(cs: Sequence[JsonConceptScheme], urn: str) -> JsonConcept:
    r = parse_item_urn(urn)
    f = [
        m
        for m in cs
        if (m.agency == r.agency and m.id == r.id and m.version == r.version)
    ]
    return [c for c in f[0].concepts if c.id == r.item_id][0]


def __get_type(repr_: JsonRepresentation) -> str:
    if repr_.enumerationFormat:
        t = repr_.enumerationFormat.textType
    elif repr_.format:
        t = repr_.format.textType
    else:
        t = "String"
    return t


def _get_representation(
    id_: str,
    local: Optional[JsonRepresentation],
    core: Optional[JsonRepresentation],
    cls: Sequence[Codelist],
    cons: Dict[str, Sequence[str]],
) -> Tuple[
    DataType,
    Optional[Facets],
    Sequence[Code],
    Optional[str],
    Optional[ArrayBoundaries],
]:
    valid = cons.get(id_, [])
    codes: Sequence[Code] = []
    dt = DataType.STRING
    facets = None
    cl_ref = None
    ab = None
    if local:
        dt = DataType(__get_type(local))
        facets = local.to_facets()
        codes = local.to_enumeration(cls, valid)
        cl_ref = local.enumeration
        ab = local.to_array_def()
    elif core:
        dt = DataType(__get_type(core))
        facets = core.to_facets()
        codes = core.to_enumeration(cls, valid)
        cl_ref = core.enumeration
        ab = core.to_array_def()
    return (dt, facets, codes, cl_ref, ab)


class JsonGroup(Struct, frozen=True):
    """SDMX-JSON payload for a group."""

    id: str
    groupDimensions: Sequence[str]


class JsonAttributeRelationship(Struct, frozen=True):
    """SDMX-JSON payload for an attribute relationship."""

    dimensions: Optional[Sequence[str]] = None
    group: Optional[str] = None

    def to_model(
        self, groups: Sequence[JsonGroup], measures: Optional[Sequence[str]]
    ) -> str:
        """Returns the attachment level."""
        if measures:
            return "O"
        elif self.dimensions:
            return ",".join(self.dimensions)
        elif self.group:
            grp = [g for g in groups if g.id == self.group]
            return ",".join(grp[0].groupDimensions)
        else:
            return "D"


class JsonDimension(Struct, frozen=True):
    """SDMX-JSON payload for a component."""

    id: str
    conceptIdentity: str
    localRepresentation: Optional[JsonRepresentation] = None

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
    ) -> Component:
        """Returns a component."""
        c = _find_concept(cs, self.conceptIdentity)
        dt, facets, codes, cl_ref, ab = _get_representation(
            self.id, self.localRepresentation, c.coreRepresentation, cls, cons
        )
        return Component(
            self.id,
            True,
            Role.DIMENSION,
            dt,
            facets,
            c.name,
            c.description,
            codes=codes,
            enum_ref=cl_ref,
            array_def=ab,
        )


class JsonAttribute(Struct, frozen=True):
    """SDMX-JSON payload for an attribute."""

    id: str
    conceptIdentity: str
    attributeRelationship: JsonAttributeRelationship
    usage: str = "optional"
    measureRelationship: Optional[Sequence[str]] = None
    localRepresentation: Optional[JsonRepresentation] = None

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
        groups: Sequence[JsonGroup],
    ) -> Component:
        """Returns a component."""
        c = _find_concept(cs, self.conceptIdentity)
        dt, facets, codes, cl_ref, ab = _get_representation(
            self.id, self.localRepresentation, c.coreRepresentation, cls, cons
        )
        req = self.usage != "optional"
        lvl = self.attributeRelationship.to_model(
            groups,
            self.measureRelationship,
        )
        return Component(
            self.id,
            req,
            Role.ATTRIBUTE,
            dt,
            facets,
            c.name,
            c.description,
            codes=codes,
            attachment_level=lvl,
            enum_ref=cl_ref,
            array_def=ab,
        )


class JsonMeasure(Struct, frozen=True):
    """SDMX-JSON payload for a measure."""

    id: str
    conceptIdentity: str
    usage: str = "optional"
    localRepresentation: Optional[JsonRepresentation] = None

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
    ) -> Component:
        """Returns a component."""
        c = _find_concept(cs, self.conceptIdentity)
        dt, facets, codes, cl_ref, ab = _get_representation(
            self.id, self.localRepresentation, c.coreRepresentation, cls, cons
        )
        req = self.usage != "optional"
        return Component(
            self.id,
            req,
            Role.MEASURE,
            dt,
            facets,
            c.name,
            c.description,
            codes=codes,
            enum_ref=cl_ref,
            array_def=ab,
        )


class JsonAttributes(Struct, frozen=True):
    """SDMX-JSON payload for the list of attributes."""

    attributes: Sequence[JsonAttribute] = ()

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
        groups: Sequence[JsonGroup],
    ) -> List[Component]:
        """Returns the list of attributes."""
        return [d.to_model(cs, cls, cons, groups) for d in self.attributes]


class JsonDimensions(Struct, frozen=True):
    """SDMX-JSON payload for the list of dimensions."""

    dimensions: Sequence[JsonDimension]
    timeDimensions: Sequence[JsonDimension] = ()

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
    ) -> List[Component]:
        """Returns the list of dimensions."""
        c = []
        c.extend([d.to_model(cs, cls, cons) for d in self.dimensions])
        c.extend([d.to_model(cs, cls, cons) for d in self.timeDimensions])
        return c


class JsonMeasures(Struct, frozen=True):
    """SDMX-JSON payload for the list of measures."""

    measures: Sequence[JsonMeasure]

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
    ) -> List[Component]:
        """Returns the list of measures."""
        return [m.to_model(cs, cls, cons) for m in self.measures]


class JsonComponents(Struct, frozen=True):
    """SDMX-JSON payload for the list of DSD components."""

    dimensionList: JsonDimensions
    measureList: Optional[JsonMeasures] = None
    attributeList: Optional[JsonAttributes] = None
    groups: Sequence[JsonGroup] = ()

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        constraints: Sequence[JsonContentConstraint],
    ) -> Components:
        """Returns the schema for this DSD."""
        comps = []
        if constraints:
            cons = constraints[0].cubeRegions[0].to_map()
        else:
            cons = {}
        comps.extend(self.dimensionList.to_model(cs, cls, cons))
        if self.measureList:
            comps.extend(self.measureList.to_model(cs, cls, cons))
        if self.attributeList:
            comps.extend(
                self.attributeList.to_model(
                    cs,
                    cls,
                    cons,
                    self.groups,
                )
            )
        return Components(comps)


class JsonDataStructure(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a DSD."""

    id: str
    name: str
    agency: str
    dataStructureComponents: JsonComponents
    description: Optional[str] = None
    version: str = "1.0"



"""Collection of SDMX-JSON schemas for dataflow queries."""

from typing import List, Optional, Sequence

from msgspec import Struct

from pysdmx.fmr.sdmx.org import JsonDataProviderScheme
from pysdmx.model import Components, DataflowInfo, DataflowRef, Organisation


class JsonDataflowRef(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a dataflow."""

    id: str
    agency: str
    name: Optional[str] = None
    description: Optional[str] = None
    version: str = "1.0"

    def to_model(self) -> DataflowRef:
        """Converts a JsonDataflowRef to a standard dataflow ref."""
        return DataflowRef(
            self.id,
            self.agency,
            self.name,
            self.description,
            self.version,
        )


class JsonDataflow(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a dataflow."""

    id: str
    name: str
    agency: str
    structure: str
    description: Optional[str] = None
    version: str = "1.0"


class JsonDataflows(Struct, frozen=True):
    """SDMX-JSON payload for the list of dataflows."""

    dataflows: Sequence[JsonDataflow]
    dataProviderSchemes: Sequence[JsonDataProviderScheme] = ()

    def __filter(
        self, df: JsonDataflow, agency: str, id_: str, version: str
    ) -> bool:
        if version != "+" and version != "latest":
            return (
                df.agency == agency and df.id == id_ and df.version == version
            )
        else:
            return df.agency == agency and df.id == id_

    def to_model(
        self, components: Components, agency: str, id_: str, version: str
    ) -> DataflowInfo:
        """Returns the requested dataflow details."""
        prvs: List[Organisation] = []
        for dps in self.dataProviderSchemes:
            prvs.extend(dps.dataProviders)
        df = list(
            filter(
                lambda df: self.__filter(df, agency, id_, version),
                self.dataflows,
            )
        )[0]
        return DataflowInfo(
            df.id,
            components,
            Organisation(df.agency),
            df.name,
            df.description,
            df.version,
            prvs,
            dsd_ref=df.structure,
        )


class JsonDataflowMessage(Struct, frozen=True):
    """SDMX-JSON payload for /dataflow queries."""

    data: JsonDataflows

    def to_model(
        self, components: Components, agency: str, id_: str, version: str
    ) -> DataflowInfo:
        """Returns the requested dataflow details."""
        return self.data.to_model(components, agency, id_, version)



"""Collection of Fusion-JSON schemas for dataflow queries."""

from typing import List, Optional, Sequence

from msgspec import Struct

from pysdmx.fmr.fusion.core import FusionString
from pysdmx.fmr.fusion.org import FusionProviderScheme
from pysdmx.model import Components, DataflowInfo, DataflowRef, Organisation


class FusionDataflowRef(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a dataflow."""

    id: str
    agency: str
    names: Sequence[FusionString]
    descriptions: Optional[Sequence[FusionString]] = None
    version: str = "1.0"

    def to_model(self) -> DataflowRef:
        """Converts a FusionDataflowRef to a standard dataflow ref."""
        return DataflowRef(
            self.id,
            self.agency,
            self.names[0].value if self.names else None,
            self.descriptions[0].value if self.descriptions else None,
            self.version,
        )


class FusionDataflow(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a dataflow."""

    id: str
    names: Sequence[FusionString]
    agency: str
    dataStructureRef: str
    descriptions: Optional[Sequence[FusionString]] = None
    version: str = "1.0"


class FusionDataflowMessage(Struct, frozen=True):
    """Fusion-JSON payload for /dataflow queries."""

    Dataflow: Sequence[FusionDataflow]
    DataProviderScheme: Sequence[FusionProviderScheme] = ()

    def __filter(
        self,
        df: FusionDataflow,
        agency: str,
        id_: str,
        version: str,
    ) -> bool:
        if version != "+" and version != "latest":
            return (
                df.agency == agency and df.id == id_ and df.version == version
            )
        else:
            return df.agency == agency and df.id == id_

    def to_model(
        self, components: Components, agency: str, id_: str, version: str
    ) -> DataflowInfo:
        """Returns the requested dataflow details."""
        prvs: List[Organisation] = []
        for dps in self.DataProviderScheme:
            prvs.extend(dps.to_model([]))
        df = list(
            filter(
                lambda df: self.__filter(df, agency, id_, version),
                self.Dataflow,
            )
        )[0]
        return DataflowInfo(
            df.id,
            components,
            Organisation(df.agency),
            df.names[0].value,
            df.descriptions[0].value if df.descriptions else None,
            df.version,
            prvs,
            dsd_ref=df.dataStructureRef,
        )



"""API for FMR readers."""

from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, List, Protocol, runtime_checkable, Sequence

from pysdmx.model import MetadataAttribute


@runtime_checkable
class Deserializer(Protocol):
    """Parses a message and return domain objects."""

    def to_model(self) -> Any:
        """Returns the domain objects."""


@dataclass
class Deserializers:
    """Collection of deserializers for a format."""

    agencies: Deserializer
    categories: Deserializer
    codes: Deserializer
    concepts: Deserializer
    dataflow: Deserializer
    providers: Deserializer
    schema: Deserializer
    hierarchy: Deserializer
    report: Deserializer
    mapping: Deserializer
    code_map: Deserializer


def _merge_attributes(
    attrs: Sequence[MetadataAttribute],
) -> Sequence[MetadataAttribute]:
    """Groups together the values of attributes with the same ID.

    The function assumes that an attribute will either have a
    value or will act as a container for other attributes. In
    case the attribute contains other attributes AND has a value,
    this function will NOT work as expected.

    Args:
        attrs: The list of attributes to be merged

    Returns:
        The list of (possibly merged) attributes
    """
    by_id: Dict[str, List[Any]] = defaultdict(list)
    sub_id = []

    for attr in attrs:
        if attr.attributes:
            sub_id.append(
                MetadataAttribute(
                    attr.id,
                    attr.value,
                    _merge_attributes(attr.attributes),
                )
            )
        else:
            by_id[attr.id].append(attr.value)

    out = []
    out.extend(sub_id)
    for k, v in by_id.items():
        val = v if len(v) > 1 else v[0]
        out.append(MetadataAttribute(k, val))
    return out



"""Model for SDMX Concepts and Concept Schemes.

Concepts are used to **describe the relevant characteristics of statistical
domains**. For example, exchanges rates might be described with concepts such
as the numerator currency, the denominator currency, the type of exchange
rates, etc.

Additional information about the rules the values for a concept must follow
can be defined using ``DataType``, ``Facets`` or enumeration (i.e. list of
``codes``).
"""

from datetime import datetime
from enum import Enum
from typing import Iterator, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.model.code import Code


class DataType(str, Enum):
    """The expected data type for a concept."""

    ALPHA = "Alpha"
    """Alphabetical characters."""
    ALPHA_NUM = "AlphaNumeric"
    """Alphabetical and numerical characters."""
    BIG_INTEGER = "BigInteger"
    """Immutable arbitrary-precision signed integer."""
    BOOLEAN = "Boolean"
    """True or False."""
    DATE = "GregorianDay"
    """A ISO 8601 date (e.g. ``2011-06-17``)."""
    DATE_TIME = "DateTime"
    """An ISO 8601 point in time."""
    DAY = "Day"
    """A day in the ISO 8601 calendar (e.g. ``---31``)."""
    DECIMAL = "Decimal"
    """Immutable arbitrary-precision signed decimal number."""
    DOUBLE = "Double"
    """A decimal number (8 bytes)."""
    FLOAT = "Float"
    """A decimal number (4 bytes)."""
    INTEGER = "Integer"
    """A whole number (4 bytes)."""
    LONG = "Long"
    """A whole number (8 bytes)."""
    MONTH = "Month"
    """A month in the ISO 8601 calendar (e.g. ``12``)."""
    MONTH_DAY = "MonthDay"
    """A month day in the ISO 8601 calendar (e.g. ``--12-31``)."""
    NUMERIC = "Numeric"
    """A numerical value (integer or decimal)."""
    PERIOD = "ObservationalTimePeriod"
    """A reporting period. The format varies with the frequency."""
    SHORT = "Short"
    """A whole number (2 bytes)."""
    STRING = "String"
    """A string (as immutable sequence of Unicode code points)."""
    TIME = "Time"
    """An ISO 8601 time (e.g. ``12:50:42``)."""
    URI = "URI"
    """A uniform resource identifier, such as a URL."""
    YEAR = "GregorianYear"
    """An ISO 8601 year (e.g. ``2000``)."""
    YEAR_MONTH = "GregorianYearMonth"
    """An ISO 8601 year and month (e.g. ``2000-01``)."""


class Facets(Struct, frozen=True, omit_defaults=True):
    """Additional information about the concept expected values.

    The facets that apply vary with the type. For example,
    ``min_mength`` and ``max_length`` are typically used with
    strings, while ``start_value`` and ``end_value`` are typically
    used with numeric sequences.
    """

    min_length: Optional[int] = None
    """The minimum length for the value (e.g. ``1`` character)."""
    max_length: Optional[int] = None
    """The maximum length for the value (e.g. ``256`` characters)."""
    min_value: Optional[float] = None
    """Used by ranges to indicate what the lower bound of the range is."""
    max_value: Optional[float] = None
    """Used by ranges to indicate what the upper bound of the range is."""
    start_value: Optional[float] = None
    """Used in a numeric sequence to indicate its starting point."""
    end_value: Optional[float] = None
    """Used in a numeric sequence to indicate its ending point."""
    interval: Optional[Union[int, float]] = None
    """The permitted interval (increment) in a sequence."""
    time_interval: Optional[str] = None
    """The permitted duration in a time sequence."""
    decimals: Optional[int] = None
    """The number of characters allowed after the decimal separator."""
    pattern: Optional[str] = None
    """A regular expression the value must match."""
    start_time: Optional[datetime] = None
    """Indicates the starting point of a sequence."""
    end_time: Optional[datetime] = None
    """Indicates the ending point of a sequence."""
    is_sequence: bool = False
    """Whether the values are intended to be ordered."""

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = []
        for k in self.__annotations__.keys():
            v = self.__getattribute__(k)
            if v:
                out.append(f"{k}={v}")
        return ", ".join(out)


class Concept(Struct, frozen=True, omit_defaults=True):
    """A concept (aka **variable**), such as frequency, reference area, etc.

    Concepts are used to **describe the relevant characteristics** of a
    statistical domain. For example, exchanges rates might be described with
    concepts such as the numerator currency, the denominator currency,
    the type of exchange rates, etc.

    Some of these concepts are expected to be useful **across statistical
    domains**. Examples of such concepts include the frequency, the
    observation status, the confidentiality, etc.

    The *codes* field indicates the expected (i.e. allowed) set of values a
    concept can take within a particular domain. In addition to
    (or instead of) a set of codes, additional details about the expected
    format may be found in the *facets* and *dtype* fields.


    Attributes:
        id: A unique identifier for the concept (e.g. FREQ).
        dtype: The concept's data type (string, number, etc.).
        facets: Additional details such as the concept's minimum length.
        name: The concept's name.
        description: Additional descriptive information about the concept.
        codes: The expected values for the concept (e.g. a list of currency
            codes).
        enum_ref: The URN of the enumeration (codelist or valuelist) from
            which the codes are taken.
    """

    id: str
    dtype: DataType = DataType.STRING
    facets: Optional[Facets] = None
    name: Optional[str] = None
    description: Optional[str] = None
    codes: Sequence[Code] = ()
    enum_ref: Optional[str] = None

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = []
        for k in self.__annotations__.keys():
            v = self.__getattribute__(k)
            if v:
                out.append(f"{k}={str(v)}")
        return ", ".join(out)


class ConceptScheme(Struct, frozen=True, omit_defaults=True):
    """An immutable collection of concepts.

    A concept scheme is **maintained by its agency**, typically, an
    organisation like the BIS, the ECB, the IMF, SDMX, etc.

    A concept scheme has an **identifier** and a **name**. It may also have a
    **description** and a **version**.

    A concept scheme is **iterable**, i.e. it can be used directly in a for
    loop.

    Attributes:
        id: The identifier for the scheme (e.g. CROSS_DOMAIN_CONCEPTS).
        name: The scheme name (e.g. "SDMX Cross Domain Concepts").
        agency: The maintainer of the scheme (e.g. SDMX).
        description: Additional descriptive information about the scheme
            (e.g. "The set of concepts in the SDMX Glossary").
        version: The scheme version (e.g. 2.0)
        concepts: The list of concepts in the scheme.
    """

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    concepts: Sequence[Concept] = ()

    def __iter__(self) -> Iterator[Concept]:
        """Return an iterator over the list of concepts."""
        yield from self.concepts

    def __len__(self) -> int:
        """Return the number of concepts in the concept scheme."""
        return len(self.concepts)

    def __getitem__(self, id_: str) -> Optional[Concept]:
        """Return the concept identified by the given ID."""
        out = list(filter(lambda concept: concept.id == id_, self.concepts))
        if len(out) == 0:
            return None
        else:
            return out[0]



"""Collection of Fusion-JSON schemas for SDMX-REST schema queries."""
from typing import Dict, List, Optional, Sequence, Tuple

from msgspec import Struct

from pysdmx.errors import ServiceError
from pysdmx.fmr.fusion.code import FusionCodelist
from pysdmx.fmr.fusion.concept import FusionConcept, FusionConceptScheme
from pysdmx.fmr.fusion.constraint import FusionContentConstraint
from pysdmx.fmr.fusion.core import FusionRepresentation, FusionString
from pysdmx.model import (
    ArrayBoundaries,
    Code,
    Component,
    Components,
    DataType,
    Facets,
    Role,
)
from pysdmx.util import parse_item_urn


def _find_concept(
    cs: Sequence[FusionConceptScheme],
    urn: str,
) -> FusionConcept:
    r = parse_item_urn(urn)
    f = [
        m
        for m in cs
        if (m.agency == r.agency and m.id == r.id and m.version == r.version)
    ]
    return [c for c in f[0].items if c.id == r.item_id][0]


def _get_representation(
    id_: str,
    repr_: Optional[FusionRepresentation],
    cls: Sequence[FusionCodelist],
    cons: Dict[str, Sequence[str]],
    c: Optional[FusionConcept],
) -> Tuple[
    DataType,
    Optional[Facets],
    Sequence[Code],
    Optional[str],
    Optional[ArrayBoundaries],
]:
    valid = cons.get(id_, [])
    codes: Sequence[Code] = []
    cl_ref = None
    ab = None
    dt = DataType.STRING
    facets = None
    codes = []
    if repr_:
        r = repr_
    elif c and c.representation:
        r = c.representation
    else:
        r = None
    if r:
        if r.textFormat:
            dt = DataType(r.textFormat.textType)
        facets = r.to_facets()
        codes = r.to_enumeration(cls, valid)
        cl_ref = r.representation
        ab = r.to_array_def()
    return (dt, facets, codes, cl_ref, ab)


class FusionGroup(Struct, frozen=True):
    """Fusion-JSON payload for a group."""

    id: str
    dimensionReferences: Sequence[str]


class FusionAttribute(Struct, frozen=True):
    """Fusion-JSON payload for an attribute."""

    id: str
    concept: str
    mandatory: bool
    attachmentLevel: str
    representation: Optional[FusionRepresentation] = None
    attachmentGroup: Optional[str] = None
    dimensionReferences: Optional[Sequence[str]] = None
    measureReferences: Optional[Sequence[str]] = None

    def __derive_level(self, groups: Sequence[FusionGroup]) -> str:
        if self.attachmentLevel == "OBSERVATION":
            return "O"
        elif self.attachmentLevel == "DATA_SET":
            return "D"
        elif self.attachmentLevel == "GROUP":
            grp = [g for g in groups if g.id == self.attachmentGroup]
            return ",".join(grp[0].dimensionReferences)
        elif self.dimensionReferences:
            return ",".join(self.dimensionReferences)
        else:
            raise ServiceError(
                500,
                "Invalid metadata",
                (
                    "Could not infer attribute attachment level. "
                    f"The attribute is {self}"
                ),
            )

    def to_model(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        cons: Dict[str, Sequence[str]],
        groups: Sequence[FusionGroup],
    ) -> Component:
        """Returns an attribute."""
        c = _find_concept(cs, self.concept)
        dt, facets, codes, cl_ref, ab = _get_representation(
            self.id, self.representation, cls, cons, c
        )
        lvl = self.__derive_level(groups)
        if c.descriptions:
            desc = c.descriptions[0].value
        else:
            desc = None
        return Component(
            self.id,
            self.mandatory,
            Role.ATTRIBUTE,
            dt,
            facets,
            c.names[0].value,
            desc,
            codes=codes,
            attachment_level=lvl,
            enum_ref=cl_ref,
            array_def=ab,
        )


class FusionAttributes(Struct, frozen=True):
    """Fusion-JSON payload for the list of attributes."""

    attributes: Sequence[FusionAttribute]

    def to_model(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        cons: Dict[str, Sequence[str]],
        groups: Sequence[FusionGroup],
    ) -> List[Component]:
        """Returns the list of attributes."""
        return [d.to_model(cs, cls, cons, groups) for d in self.attributes]


class FusionDimension(Struct, frozen=True):
    """Fusion-JSON payload for a dimension."""

    id: str
    concept: str
    representation: Optional[FusionRepresentation] = None

    def to_model(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        cons: Dict[str, Sequence[str]],
    ) -> Component:
        """Returns a dimension."""
        c = _find_concept(cs, self.concept)
        dt, facets, codes, cl_ref, ab = _get_representation(
            self.id, self.representation, cls, cons, c
        )
        if c.descriptions:
            desc = c.descriptions[0].value
        else:
            desc = None
        return Component(
            self.id,
            True,
            Role.DIMENSION,
            dt,
            facets,
            c.names[0].value,
            desc,
            codes=codes,
            enum_ref=cl_ref,
            array_def=ab,
        )


class FusionDimensions(Struct, frozen=True):
    """Fusion-JSON payload for the list of dimensions."""

    dimensions: Sequence[FusionDimension]

    def to_model(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        cons: Dict[str, Sequence[str]],
    ) -> List[Component]:
        """Returns the list of dimensions."""
        return [d.to_model(cs, cls, cons) for d in self.dimensions]


class FusionMeasure(Struct, frozen=True):
    """Fusion-JSON payload for a measure."""

    id: str
    concept: str
    mandatory: bool
    representation: Optional[FusionRepresentation] = None

    def to_model(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        cons: Dict[str, Sequence[str]],
    ) -> Component:
        """Returns a measure."""
        c = _find_concept(cs, self.concept)
        dt, facets, codes, cl_ref, ab = _get_representation(
            self.id, self.representation, cls, cons, c
        )
        if c.descriptions:
            desc = c.descriptions[0].value
        else:
            desc = None
        return Component(
            self.id,
            self.mandatory,
            Role.MEASURE,
            dt,
            facets,
            c.names[0].value,
            desc,
            codes=codes,
            enum_ref=cl_ref,
            array_def=ab,
        )


class FusionDataStructure(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a DSD."""

    id: str
    names: Sequence[FusionString]
    agency: str
    dimensionList: FusionDimensions
    measures: Sequence[FusionMeasure] = ()
    attributeList: Optional[FusionAttributes] = None
    groups: Sequence[FusionGroup] = ()
    description: Optional[Sequence[FusionString]] = None
    version: str = "1.0"

    def get_components(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        constraints: Sequence[FusionContentConstraint],
    ) -> Components:
        """Returns the schema for this DSD."""
        comps = []
        if constraints:
            cons = constraints[0].to_map()
        else:
            cons = {}
        comps.extend(self.dimensionList.to_model(cs, cls, cons))
        if self.measures:
            comps.extend([m.to_model(cs, cls, cons) for m in self.measures])
        if self.attributeList:
            comps.extend(
                self.attributeList.to_model(
                    cs,
                    cls,
                    cons,
                    self.groups,
                )
            )
        return Components(comps)



"""Collection of Fusion-JSON schemas for the metadata queries."""

from pysdmx.fmr.fusion.category import FusionCategorySchemeMessage
from pysdmx.fmr.fusion.code import (
    FusionCodelistMessage,
    FusionHierarchyMessage,
)
from pysdmx.fmr.fusion.concept import FusionConcepSchemeMessage
from pysdmx.fmr.fusion.dataflow import FusionDataflowMessage
from pysdmx.fmr.fusion.map import (
    FusionMappingMessage,
    FusionRepresentationMapMessage,
)
from pysdmx.fmr.fusion.org import FusionAgencyMessage, FusionProviderMessage
from pysdmx.fmr.fusion.report import FusionMetadataMessage
from pysdmx.fmr.fusion.schema import FusionSchemaMessage
from pysdmx.fmr.reader import Deserializers

deserializers = Deserializers(
    agencies=FusionAgencyMessage,  # type: ignore[arg-type]
    categories=FusionCategorySchemeMessage,  # type: ignore[arg-type]
    codes=FusionCodelistMessage,  # type: ignore[arg-type]
    concepts=FusionConcepSchemeMessage,  # type: ignore[arg-type]
    dataflow=FusionDataflowMessage,  # type: ignore[arg-type]
    providers=FusionProviderMessage,  # type: ignore[arg-type]
    schema=FusionSchemaMessage,  # type: ignore[arg-type]
    hierarchy=FusionHierarchyMessage,  # type: ignore[arg-type]
    report=FusionMetadataMessage,  # type: ignore[arg-type]
    mapping=FusionMappingMessage,  # type: ignore[arg-type]
    code_map=FusionRepresentationMapMessage,  # type: ignore[arg-type]
)



"""Collection of Fusion-JSON schemas for common artefacts."""

from datetime import datetime
from typing import Any, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.model import ArrayBoundaries, Code, Facets
from pysdmx.util import find_by_urn


class FusionAnnotation(Struct, frozen=True):
    """Fusion-JSON payload for annotations."""

    title: str
    type: str


class FusionString(Struct, frozen=True):
    """Fusion-JSON payload for an international string."""

    locale: str
    value: str


class FusionTextFormat(Struct, frozen=True):
    """Fusion-JSON payload for TextFormat."""

    textType: str
    minLength: Optional[int] = None
    maxLength: Optional[int] = None
    minValue: Optional[Union[int, float]] = None
    maxValue: Optional[Union[int, float]] = None
    startValue: Optional[Union[int, float]] = None
    endValue: Optional[Union[int, float]] = None
    decimals: Optional[int] = None
    pattern: Optional[str] = None
    startTime: Optional[datetime] = None
    endTime: Optional[datetime] = None
    isSequence: bool = False


class FusionRepresentation(Struct, frozen=True):
    """Fusion-JSON payload for core representation."""

    textFormat: Optional[FusionTextFormat] = None
    representation: Optional[str] = None
    minOccurs: Optional[int] = None
    maxOccurs: Optional[int] = None

    def to_facets(self) -> Optional[Facets]:
        """Return a Facets domain object."""
        if self.textFormat and (
            self.textFormat.minLength
            or self.textFormat.maxLength
            or self.textFormat.isSequence
            or self.textFormat.minValue
            or self.textFormat.maxValue
            or self.textFormat.startValue
            or self.textFormat.endValue
            or self.textFormat.decimals
            or self.textFormat.pattern
            or self.textFormat.startTime
            or self.textFormat.endTime
        ):
            return Facets(
                min_length=self.textFormat.minLength,
                max_length=self.textFormat.maxLength,
                is_sequence=self.textFormat.isSequence,
                min_value=self.textFormat.minValue,
                max_value=self.textFormat.maxValue,
                start_value=self.textFormat.startValue,
                end_value=self.textFormat.endValue,
                decimals=self.textFormat.decimals,
                pattern=self.textFormat.pattern,
                start_time=self.textFormat.startTime,
                end_time=self.textFormat.endTime,
            )
        else:
            return None

    def to_enumeration(
        self,
        codelists: Sequence[Any],
        valid: Sequence[str],
    ) -> Sequence[Code]:
        """Returns the list of codes allowed for this component."""
        codes = []
        if self.representation:
            a = find_by_urn(codelists, self.representation).items
            codes = [c.to_model() for c in a if not valid or c.id in valid]
        return codes

    def to_array_def(self) -> Optional[ArrayBoundaries]:
        """Returns the array boundaries, if any."""
        if self.minOccurs or self.maxOccurs:
            m = self.minOccurs if self.minOccurs is not None else 0
            return ArrayBoundaries(m, self.maxOccurs)
        else:
            return None



"""Collection of SDMX-JSON schemas for content constraints."""

from typing import Dict, Sequence

from msgspec import Struct


class JsonValue(Struct, frozen=True):
    """SDMX-JSON payload for an allowed value."""

    value: str


class JsonKeyValue(Struct, frozen=True):
    """SDMX-JSON payload for the list of allowed values per component."""

    id: str
    values: Sequence[JsonValue]

    def to_model(self) -> Sequence[str]:
        """Returns the requested list of values."""
        return [v.value for v in self.values]


class JsonCubeRegion(Struct, frozen=True):
    """SDMX-JSON payload for a cube region."""

    keyValues: Sequence[JsonKeyValue]

    def to_map(self) -> Dict[str, Sequence[str]]:
        """Gets the list of allowed values for a component."""
        return {kv.id: kv.to_model() for kv in self.keyValues}


class JsonContentConstraint(Struct, frozen=True):
    """SDMX-JSON payload for a content constraint."""

    cubeRegions: Sequence[JsonCubeRegion]



"""Collection of Fusion-JSON schemas for reference metadata queries."""

from typing import Sequence

from msgspec import Struct

from pysdmx.fmr.fusion.core import FusionString
from pysdmx.fmr.reader import _merge_attributes
from pysdmx.model import MetadataAttribute, MetadataReport


class FusionMetadataReport(Struct, frozen=True):
    """Fusion-JSON payload for a metadata report."""

    id: str
    names: Sequence[FusionString]
    metadataflow: str
    targets: Sequence[str]
    attributes: Sequence[MetadataAttribute]


class FusionMetadataSets(Struct, frozen=True):
    """Fusion-JSON payload for the list of metadata sets."""

    metadatasets: Sequence[FusionMetadataReport]


class FusionMetadataMessage(Struct, frozen=True):
    """Fusion-JSON payload for /metadata queries."""

    data: FusionMetadataSets

    def to_model(self) -> MetadataReport:
        """Returns the requested metadata report."""
        r = self.data.metadatasets[0]
        attrs = _merge_attributes(r.attributes)
        return MetadataReport(
            r.id, r.names[0].value, r.metadataflow, r.targets, attrs
        )



"""Collection of Fusion-JSON schemas for codes and codelists."""

from datetime import datetime, timedelta, timezone as tz
from typing import Optional, Sequence, Tuple

from msgspec import Struct

from pysdmx.fmr.fusion.core import FusionAnnotation, FusionString
from pysdmx.model import (
    Code,
    Codelist as CL,
    HierarchicalCode,
    Hierarchy as HCL,
)
from pysdmx.util import parse_item_urn


class FusionCode(Struct, frozen=True):
    """Fusion-JSON payload for codes."""

    id: str
    annotations: Sequence[FusionAnnotation] = ()
    names: Sequence[FusionString] = ()
    descriptions: Sequence[FusionString] = ()

    def __handle_date(self, datestr: str) -> datetime:
        return datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%S%z")

    def __get_val(
        self, a: FusionAnnotation
    ) -> Tuple[Optional[datetime], Optional[datetime]]:
        vals = a.title.split("/")
        if a.title.startswith("/"):
            return (None, self.__handle_date(vals[1]))
        else:
            valid_from = self.__handle_date(vals[0])
            valid_to = self.__handle_date(vals[1]) if vals[1] else None
            return (valid_from, valid_to)

    def to_model(self) -> Code:
        """Converts a FusionCode to a standard code."""
        vp = [a for a in self.annotations if a.type == "FR_VALIDITY_PERIOD"]
        vf, vt = self.__get_val(vp[0]) if vp else (None, None)
        return Code(
            self.id,
            self.names[0].value,
            self.descriptions[0].value if self.descriptions else None,
            vf,
            vt,
        )


class FusionCodelist(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a codelist."""

    id: str
    names: Sequence[FusionString]
    agency: str
    descriptions: Sequence[FusionString] = ()
    version: str = "1.0"
    items: Sequence[FusionCode] = ()

    def to_model(self) -> CL:
        """Converts a JsonCodelist to a standard codelist."""
        return CL(
            self.id,
            self.names[0].value,
            self.agency,
            self.descriptions[0].value if self.descriptions else None,
            self.version,
            [i.to_model() for i in self.items],
        )


class FusionCodelistMessage(Struct, frozen=True):
    """Fusion-JSON payload for /codelist queries."""

    Codelist: Sequence[FusionCodelist] = ()
    ValueList: Sequence[FusionCodelist] = ()

    def to_model(self) -> CL:
        """Returns the requested codelist."""
        if self.Codelist:
            return self.Codelist[0].to_model()
        else:
            return self.ValueList[0].to_model()


class FusionHierarchicalCode(Struct, frozen=True):
    """Fusion-JSON payload for hierarchical codes."""

    code: str
    validFrom: Optional[int] = None
    validTo: Optional[int] = None
    codes: Sequence["FusionHierarchicalCode"] = ()

    def __find_code(self, codelists: Sequence[CL], urn: str) -> Code:
        r = parse_item_urn(urn)
        f = [
            c
            for c in codelists
            if (
                c.agency == r.agency
                and c.id == r.id
                and c.version == r.version
            )
        ]
        return [c for c in f[0].codes if c.id == r.item_id][0]

    def __convert_epoch(self, epoch: int) -> datetime:
        if epoch < 0:
            return datetime(1970, 1, 1, tzinfo=tz.utc) + timedelta(
                milliseconds=epoch
            )
        else:
            return datetime.fromtimestamp(epoch / 1000, tz.utc)

    def to_model(self, codelists: Sequence[CL]) -> HierarchicalCode:
        """Converts a FusionHierarchicalCode to a hierachical code."""
        code = self.__find_code(codelists, self.code)
        rvf = self.__convert_epoch(self.validFrom) if self.validFrom else None
        rvt = self.__convert_epoch(self.validTo) if self.validTo else None
        codes = [c.to_model(codelists) for c in self.codes]
        return HierarchicalCode(
            code.id,
            code.name,
            code.description,
            code.valid_from,
            code.valid_to,
            rvf,
            rvt,
            codes,
        )


class FusionHierarchy(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a hierarchy."""

    id: str
    names: Sequence[FusionString]
    agency: str
    descriptions: Sequence[FusionString] = ()
    version: str = "1.0"
    codes: Sequence[FusionHierarchicalCode] = ()

    def to_model(self, codelists: Sequence[CL]) -> HCL:
        """Converts a FusionHierarchy to a standard hierarchy."""
        return HCL(
            self.id,
            self.names[0].value,
            self.agency,
            self.descriptions[0].value if self.descriptions else None,
            self.version,
            [i.to_model(codelists) for i in self.codes],
        )


class FusionHierarchyMessage(Struct, frozen=True):
    """Fusion-JSON payload for /hierarchy queries."""

    Codelist: Sequence[FusionCodelist] = ()
    Hierarchy: Sequence[FusionHierarchy] = ()

    def to_model(self) -> HCL:
        """Returns the requested hierarchy."""
        cls = [cl.to_model() for cl in self.Codelist]
        return self.Hierarchy[0].to_model(cls)



"""Model for SDMX Codes and Codelists.

Two different types of collections are provided:

- Codelist: A flat structure, i.e. a structure where all codes are at the same
  level. In SDMX terms, this structure can be fed from information stored
  as a *codelist* or a *value list*.
- Hierarchy: A structure where codes can be organized in a hierarchical
  fashion, like a list of country groups and their respective countries.

This is an example of what is meant with pysdmx being opinionated.
While SDMX allows two different types of hierarchies (simple hierarchies
expressed as codelists, and more complex hierarchies expressed as SDMX
hierarchies), pysdmx expects codelists and value lists to be flat (which
seems to be the vast majority of the cases), thereby leaving the
representation of hierarchical relationships to hierarchies only.
"""

from datetime import datetime
from typing import Iterator, Optional, Sequence

from msgspec import Struct


class Code(Struct, frozen=True, omit_defaults=True):
    """A code, such as a country code in the list of ISO 3166 codes.

    Codes may have business validity information.

    Attributes:
        id: The identifier for the code (e.g. UY).
        name: The code's name (e.g. Uruguay).
        description: Additional descriptive information about the code.
        valid_from: Start of the code's validity period.
        valid_to: End of the code's validity period.
    """

    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    valid_from: Optional[datetime] = None
    valid_to: Optional[datetime] = None

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = self.id
        if self.name:
            out = f"{out} ({self.name})"
        return out


class Codelist(Struct, frozen=True, omit_defaults=True):
    """An immutable collection of codes, such as the ISO 3166 country codes.

    A codelist is **maintained by its agency**, typically, an organisation
    like the BIS, the ECB, the IMF, etc.

    A codelist has an **identifier** and a **name**. It may also have a
    **description** and **business validity information**.

    A codelist is **iterable**, i.e. it can be used as is in a for loop.

    A codelist represents a flat list of codes, without any hierarchical
    relationship between them. Both SDMX codelists and SDMX value lists
    are represented as a Codelist class.

    Attributes:
        id: The identifier for the codelist (e.g. CL_FREQ).
        name: The codelist name (e.g. "Frequency codelist").
        agency: The maintainer of the codelist (e.g. SDMX).
        description: Additional descriptive information about the codelist
            (e.g. "This codelist provides a set of values indicating the
            frequency of the data").
        version: The codelist version (e.g. 2.0.42)
        codes: The list of codes in the codelist.
    """

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    codes: Sequence[Code] = ()

    def __iter__(self) -> Iterator[Code]:
        """Return an iterator over the list of codes."""
        yield from self.codes

    def __len__(self) -> int:
        """Return the number of codes in the codelist."""
        return len(self.codes)

    def __getitem__(self, id_: str) -> Optional[Code]:
        """Return the code identified by the supplied ID."""
        out = list(filter(lambda code: code.id == id_, self.codes))
        if len(out) == 0:
            return None
        else:
            return out[0]


class HierarchicalCode(Struct, frozen=True, omit_defaults=True):
    """A code, as used in a hierarchy.

    Hierachical codes may contain other codes.

    As codes, hierarchial codes may have business validity information.
    In addition, they may also have business validity associated with
    their relationship in the hierarchy.

    For example, let's imagine a hierarchy representing country groups.
    A country might have a valid_from property, with the date that
    country declared its independence. In addition, it may also
    have a rel_valid_from property, with the date the country joined
    that particular group (e.g. European Union).

    Attributes:
        id: The identifier for the code (e.g. UY).
        name: The code's name (e.g. Uruguay).
        description: Additional descriptive information about the code.
        valid_from: Start of the code's validity period.
        valid_to: End of the code's validity period.
        rel_valid_from: Start of the hierarchical relationship validity.
        rel_valid_to: End of the hierarchical relationship validity.
        codes: The child codes.
    """

    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    valid_from: Optional[datetime] = None
    valid_to: Optional[datetime] = None
    rel_valid_from: Optional[datetime] = None
    rel_valid_to: Optional[datetime] = None
    codes: Sequence["HierarchicalCode"] = ()

    def __iter__(self) -> Iterator["HierarchicalCode"]:
        """Return an iterator over the list of codes."""
        yield from self.codes

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = self.id
        if self.name:
            out = f"{out} ({self.name})"
        return out


class Hierarchy(Struct, frozen=True, omit_defaults=True):
    """An immutable collection of codes, organized hierarchically.

    A hierarchy is **maintained by its agency**, typically, an organisation
    like the BIS, the ECB, the IMF, etc.

    A hierarchy has an **identifier** and a **name**. It may also have a
    **description** and **business validity information**.

    A hierarchy is **iterable**, i.e. it can be used as is in a for loop.

    Attributes:
        id: The identifier for the hierarchy (e.g. AREA).
        name: The hierarchy name (e.g. "Country groups and their composition").
        agency: The maintainer of the hierarchy (e.g. SDMX).
        description: Additional descriptive information about the hierarchy
            (e.g. "This hierarchy provides a set of country groups and their
            respective composition").
        version: The hierarchy version (e.g. 2.0.42)
        codes: The list of codes in the hierarchy.
    """

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    codes: Sequence[HierarchicalCode] = ()

    def __iter__(self) -> Iterator[HierarchicalCode]:
        """Return an iterator over the list of codes."""
        yield from self.codes

    def __len__(self) -> int:
        """Return the number of codes in the hierarchy."""
        return self.__get_count(self.codes)

    def __getitem__(self, id_: str) -> Optional[HierarchicalCode]:
        """Return the code identified by the supplied ID."""
        return self.__extract_code(self.codes, id_)

    def __get_count(self, codes: Sequence[HierarchicalCode]) -> int:
        """Return the number of codes at all levels."""
        count = len(codes)
        for code in codes:
            if code.codes:
                count += self.__get_count(code.codes)
        return count

    def __extract_code(
        self, codes: Sequence[HierarchicalCode], id_: str
    ) -> Optional[HierarchicalCode]:
        if "." in id_:
            ids = id_.split(".")
            out = list(filter(lambda c: c.id == ids[0], codes))
            if out:
                pkey = ".".join(ids[1:])
                return self.__extract_code(out[0].codes, pkey)
        else:
            out = list(filter(lambda c: c.id == id_, codes))
            if out:
                return out[0]
        return None



"""Collection of SDMX-JSON schemas for the metadata queries."""

from pysdmx.fmr.reader import Deserializers
from pysdmx.fmr.sdmx.category import JsonCategorySchemeMessage
from pysdmx.fmr.sdmx.code import JsonCodelistMessage, JsonHierarchyMessage
from pysdmx.fmr.sdmx.concept import JsonConcepSchemeMessage
from pysdmx.fmr.sdmx.dataflow import JsonDataflowMessage
from pysdmx.fmr.sdmx.map import (
    JsonMappingMessage,
    JsonRepresentationMapMessage,
)
from pysdmx.fmr.sdmx.org import JsonAgencyMessage, JsonProviderMessage
from pysdmx.fmr.sdmx.report import JsonMetadataMessage
from pysdmx.fmr.sdmx.schema import JsonSchemaMessage

deserializers = Deserializers(
    agencies=JsonAgencyMessage,  # type: ignore[arg-type]
    categories=JsonCategorySchemeMessage,  # type: ignore[arg-type]
    codes=JsonCodelistMessage,  # type: ignore[arg-type]
    concepts=JsonConcepSchemeMessage,  # type: ignore[arg-type]
    dataflow=JsonDataflowMessage,  # type: ignore[arg-type]
    providers=JsonProviderMessage,  # type: ignore[arg-type]
    schema=JsonSchemaMessage,  # type: ignore[arg-type]
    hierarchy=JsonHierarchyMessage,  # type: ignore[arg-type]
    report=JsonMetadataMessage,  # type: ignore[arg-type]
    mapping=JsonMappingMessage,  # type: ignore[arg-type]
    code_map=JsonRepresentationMapMessage,  # type: ignore[arg-type]
)



"""Collection of utility functions."""

import re
from typing import Any, Sequence

from msgspec import Struct

from pysdmx.errors import NotFound


class Reference(Struct, frozen=True):
    """The coordinates of an SDMX maintainable artefact.

    Attributes:
        sdmx_type: The type of SDMX artefact (``codelist``, etc.)
        agency: The maintainer of the artefact (e.g. ``BIS``, ``SDMX``, etc.)
        id: The artefact ID (e.g. ``CL_FREQ``)
        version: The artefact version (e.g. ``1.0.0``)
    """

    sdmx_type: str
    agency: str
    id: str
    version: str


class ItemReference(Struct, frozen=True):
    """The coordinates of an SDMX non-nested item.

    Attributes:
        sdmx_type: The type of SDMX artefact (``concept``, etc.)
        agency: The maintainer of the artefact (e.g. ``BIS``, etc.)
        id: The maintainable ID (e.g. ``CL_FREQ``)
        version: The artefact version (e.g. ``1.0.0``)
        item_id: The item ID (e.g. ``A``)
    """

    sdmx_type: str
    agency: str
    id: str
    version: str
    item_id: str


maintainable_urn_pattern = re.compile(r"^.*\.(.*)=(.*):(.*)\((.*)\)$")
item_urn_pattern = re.compile(r"^.*\.(.*)=(.*):(.*)\((.*)\)\.(.*)$")


def parse_urn(urn: str) -> Reference:
    """Parses an SDMX urn and returns an object with the details."""
    m = re.match(maintainable_urn_pattern, urn)
    if m:
        return Reference(
            sdmx_type=m.group(1),
            agency=m.group(2),
            id=m.group(3),
            version=m.group(4),
        )
    else:
        raise NotFound(
            404,
            "Not found",
            f"{urn} does not match {maintainable_urn_pattern}",
        )


def parse_item_urn(urn: str) -> ItemReference:
    """Parses an SDMX item urn and returns an object with the details."""
    m = re.match(item_urn_pattern, urn)
    if m:
        return ItemReference(
            sdmx_type=m.group(1),
            agency=m.group(2),
            id=m.group(3),
            version=m.group(4),
            item_id=m.group(5),
        )
    else:
        raise NotFound(
            404, "Not found", f"{urn} does not match {item_urn_pattern}."
        )


def find_by_urn(artefacts: Sequence[Any], urn: str) -> Any:
    """Returns the maintainable artefact matching the supplied urn."""
    r = parse_urn(urn)
    f = [
        a
        for a in artefacts
        if (a.agency == r.agency and a.id == r.id and a.version == r.version)
    ]
    if f:
        return f[0]
    else:
        urns = [f"{a.agency}:{a.id}({a.version})" for a in artefacts]
        raise NotFound(
            404,
            "Not found",
            (
                f"Could not find an artefact matching the following URN: "
                f"{urn}. The artefacts received were: {urns}."
            ),
        )



"""Collection of SDMX-JSON schemas for provision agreements."""

from msgspec import Struct


class JsonProvisionAgreement(Struct, frozen=True):
    """SDMX-JSON payload for a provision agreement."""

    structureUsage: str
    dataProvider: str
